/* GFCUDA Driver

This program will load a GF256 multiplication table from file, test cases of 
various sizes, run the tests using the provided cpu inpementation and your
team's gpu implementation, and compare the return values for correctness and 
times for runetime improvement.  

DO NOT MODIFY THIS FILE!!!

Usage:
./a.out

*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#define GFSIZE 256

typedef unsigned char UINT8;

int gfmmm_cpu(UINT8 *mul_table, UINT8 *A, UINT8 *B, UINT8 *C, int n);
int gfmmm_gpu(UINT8 *mul_table, UINT8 *A, UINT8 *B, UINT8 *C, int n);

int main (int argc, char* argv[]){
  int i, j, size;
  
  /* Allocate space for the multiplication table */
  UINT8* mul_table = malloc(GFSIZE * GFSIZE * sizeof(UINT8));
  
  /* Read in the multiplicatin table from the file "gf_mul.tbl" */
  FILE* gf_file = fopen("gf_mul.tbl","r");
  for (i=0; i < GFSIZE; i++){
    for (j=0; j < GFSIZE; j++){
      fscanf(gf_file, "%c ", mul_table + i*GFSIZE + j);
    }
    fscanf(gf_file, "\n");
  }
  fclose(gf_file);
  printf("Read GF256 Multiplication Table \n");
  
  /* Pointers to matrices */
  UINT8 *A, *B, *C_cpu, *C_gpu;
  
  /* Set up random generator */
  srand(time(NULL));
  
  /* Setup clock */
  clock_t start, end;
  float time_gpu, time_cpu;
  
  /* Setup diff-counter */
  int diff = 0;
  
  /* Loop over a bunch of test sizes */
  for (size = 256; size <= 4096; size *= 2){
    printf("========================\n");
    printf(" Size: %d \n", size);
    
    /* Allocate space for matrices */
    A = (UINT8*) malloc(size * size * sizeof(UINT8));
    B = (UINT8*) malloc(size * size * sizeof(UINT8));
    C_cpu = (UINT8*) malloc(size * size * sizeof(UINT8));
    C_gpu = (UINT8*) malloc(size * size * sizeof(UINT8));
    
    for (i = 0; i < size; i++){
      for (j = 0; j < size; j++){
        A[i*size + j] = rand() % GFSIZE;
      }
    }
    
    for (i = 0; i < size; i++){
      for (j = 0; j < size; j++){
        B[i*size + j] = rand() % GFSIZE;
      }
    }
    
    /* Time the evaluation of gfmmm_gpu */
    start = clock();
    gfmmm_gpu(mul_table, A, B, C_gpu, size);
    end = clock();
    time_gpu = ((float) (clock() - start))/ CLOCKS_PER_SEC;
    printf(" GPU evaluation took %f seconds\n", time_gpu);
    
    /* Time the evaluation of gfmmm_cpu */
    start = clock();
    gfmmm_cpu(mul_table, A, B, C_cpu, size);
    end = clock();
    time_cpu = ((float) (clock() - start))/ CLOCKS_PER_SEC;
    printf(" CPU evaluation took %f seconds\n", time_cpu);
        
    /* Count the differences between the results */
    diff = 0;
    for (i = 0; i < size; i++){
      for (j = 0; j < size; j++){
        if (*(C_cpu + i*size + j) != *(C_gpu + i*size + j))
          diff++;
      }
    }
    printf(" Found %d differences in CPU/GPU output\n", diff);
    
    /* free matrices */
    free(A);
    free(B);
    free(C_cpu);
    free(C_gpu);
  }
}  


